#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include "py/compile.h"
#include "py/runtime.h"
#include "py/repl.h"
#include "py/gc.h"
#include "py/mperrno.h"
#include "py/mphal.h"
#include "shared/runtime/pyexec.h"
#include "k_csdk.h"

#if MICROPY_ENABLE_COMPILER
void do_str(const char *src, mp_parse_input_kind_t input_kind) {
    nlr_buf_t nlr;
    if (nlr_push(&nlr) == 0) {
        mp_lexer_t *lex = mp_lexer_new_from_str_len(MP_QSTR__lt_stdin_gt_, src, strlen(src), 0);
        qstr source_name = lex->source_name;
        mp_parse_tree_t parse_tree = mp_parse(lex, input_kind);
        mp_obj_t module_fun = mp_compile(&parse_tree, source_name, true);
        mp_call_function_0(module_fun);
        nlr_pop();
    } else {
        // uncaught exception
        mp_obj_print_exception(&mp_plat_print, (mp_obj_t)nlr.ret_val);
    }
}
#endif

static char *stack_top;
#if MICROPY_ENABLE_GC
static char * heap = (char *) 0xD40000+320*240;
#endif


#define FORCED_EXIT (0x100)
// If exc is SystemExit, return value where FORCED_EXIT bit set,
// and lower 8 bits are SystemExit value. For all other exceptions,
// return 1.
int handle_uncaught_exception(mp_obj_base_t *exc) {
  // check for SystemExit
  if (mp_obj_is_subclass_fast(MP_OBJ_FROM_PTR(exc->type), MP_OBJ_FROM_PTR(&mp_type_SystemExit))) {
    // None is an exit value of 0; an int is its value; anything else is 1
    mp_obj_t exit_val = mp_obj_exception_get_value(MP_OBJ_FROM_PTR(exc));
    mp_int_t val = 0;
    if (exit_val != mp_const_none && !mp_obj_get_int_maybe(exit_val, &val)) {
      val = 1;
    }
    return FORCED_EXIT | (val & 255);
  }
  
  // Report all other exceptions
  return 1;
}

#define LEX_SRC_STR (1)
#define LEX_SRC_VSTR (2)
#define LEX_SRC_FILENAME (3)
#define LEX_SRC_STDIN (4)

// Returns standard error codes: 0 for success, 1 for all other errors,
// except if FORCED_EXIT bit is set then script raised SystemExit and the
// value of the exit is in the lower 8 bits of the return value
int execute_from_lexer(int source_kind, const void *source, mp_parse_input_kind_t input_kind, bool is_repl) {
  mp_hal_stdout_tx_strn(source,strlen(source));  
  int c=mp_hal_stdin_rx_chr();
  nlr_buf_t nlr;
  if (nlr_push(&nlr) == 0) {
    // create lexer based on source kind
    mp_lexer_t *lex;
    if (source_kind == LEX_SRC_STR) {
      const char *line = source;
      lex = mp_lexer_new_from_str_len(MP_QSTR__lt_stdin_gt_, line, strlen(line), false);
    } 
    dbg_printf("lexer done\n");
    qstr source_name = lex->source_name;
    
#if MICROPY_PY___FILE__
    if (input_kind == MP_PARSE_FILE_INPUT) {
      mp_store_global(MP_QSTR___file__, MP_OBJ_NEW_QSTR(source_name));
    }
#endif
    // parser_errorline=0; parser_errorcol=0;
    mp_parse_tree_t parse_tree = mp_parse(lex, input_kind);
    dbg_printf("parse done\n");
    
    mp_obj_t module_fun = mp_compile(&parse_tree, source_name, is_repl);
    dbg_printf("compile done\n");
    // execute it
    mp_call_function_0(module_fun);
    dbg_printf("call done\n");
    
    nlr_pop();
    return 0;
  } else {
    // uncaught exception
    dbg_printf("exception\n");
    return handle_uncaught_exception(nlr.ret_val);
  }
}

int micropy_eval(const char * line){
  return execute_from_lexer(LEX_SRC_STR, line, MP_PARSE_FILE_INPUT, true);
}


int main(int argc, char **argv) {
  sdk_init();
  int stack_dummy;
  stack_top = (char *)&stack_dummy;
  
#if MICROPY_ENABLE_GC
  gc_init(heap, heap + 319*240);
#endif
  mp_init();
  dbg_printf("mp_init done\n");
  micropy_eval("1+.5");
  mp_deinit();
  sdk_end();
  return 0;
}

#if MICROPY_ENABLE_GC
void gc_collect(void) {
    // WARNING: This gc_collect implementation doesn't try to get root
    // pointers from CPU registers, and thus may function incorrectly.
    void *dummy;
    gc_collect_start();
    gc_collect_root(&dummy, ((mp_uint_t)stack_top - (mp_uint_t)&dummy) / sizeof(mp_uint_t));
    gc_collect_end();
    gc_dump_info();
}
#endif

mp_lexer_t *mp_lexer_new_from_file(const char *filename) {
    mp_raise_OSError(MP_ENOENT);
}

mp_import_stat_t mp_import_stat(const char *path) {
    return MP_IMPORT_STAT_NO_EXIST;
}

mp_obj_t mp_builtin_open(size_t n_args, const mp_obj_t *args, mp_map_t *kwargs) {
    return mp_const_none;
}
MP_DEFINE_CONST_FUN_OBJ_KW(mp_builtin_open_obj, 1, mp_builtin_open);

void nlr_jump_fail(void *val) {
    while (1) {
        ;
    }
}

void NORETURN __fatal_error(const char *msg) {
    while (1) {
        ;
    }
}
